@@ -17,11 +17,11 @@ FCGI::Buffer - Verify, Cache and Optimise FCGI Output
 
 =head1 VERSION
 
-Version 0.04
+Version 0.05
 
 =cut
 
-our $VERSION = '0.04';
+our $VERSION = '0.05';
 
 =head1 SYNOPSIS
 
@@ -73,6 +73,9 @@ But that's simple:
 	# ...
     }
 
+To temporarily prevent the use of server-side caches, for example whilst
+debugging before publishing a code change, set the NO_CACHE environment variable
+to any non-zero value.
 If you get errors about Wide characters in print it means that you've
 forgotten to emit pure HTML on non-ascii characters.
 See L<HTML::Entities>.
@@ -99,11 +102,10 @@ sub new {
 	return unless($class);
 
 	my $buf = IO::String->new();
-	my $old_buf = select($buf);
 
 	my $rc = {
 		buf => $buf,
-		old_buf => $old_buf,
+		old_buf => select($buf),
 		generate_304 => 1,
 		generate_last_modified => 1,
 		compress_content => 1,
@@ -161,15 +163,7 @@ sub DESTROY {
 	}
 
 	if($headers) {
-		foreach my $header (split(/\r?\n/, $headers)) {
-			my ($header_name, $header_value) = split /\:\s*/, $header, 2;
-			if (lc($header_name) eq 'content-type') {
-				my @content_type;
-				@content_type = split /\//, $header_value, 2;
-				$self->{content_type} = \@content_type;
-				last;
-			}
-		}
+		$self->_set_content_type($headers);
 	}
 
 	if(defined($self->{body}) && ($self->{body} eq '')) {
@@ -187,14 +181,22 @@ sub DESTROY {
 				# require HTML::Clean;
 				require HTML::Packer;	# Overkill using HTML::Clean and HTML::Packer...
 
+				if($self->{'logger'}) {
+					$self->{'logger'}->trace('Packer');
+				}
+
 				my $oldlength = length($self->{body});
 				my $newlength;
 
-				while(1) {
-					$self->{body} = $self->_optimise_content();
-					$newlength = length($self->{body});
-					last if ($newlength >= $oldlength);
-					$oldlength = $newlength;
+				if($self->{optimise_content} == 1) {
+					$self->_optimise_content();
+				} else {
+					while(1) {
+						$self->_optimise_content();
+						$newlength = length($self->{body});
+						last if ($newlength >= $oldlength);
+						$oldlength = $newlength;
+					}
 				}
 
 				# If we're on http://www.example.com and have a link
@@ -263,6 +265,9 @@ sub DESTROY {
 				require HTML::Lint;
 				HTML::Lint->import;
 
+				if($self->{'logger'}) {
+					$self->{'logger'}->trace('Lint');
+				}
 				my $lint = HTML::Lint->new();
 				$lint->parse($self->{body});
 
@@ -319,7 +324,7 @@ sub DESTROY {
 	my $encoding = $self->_should_gzip();
 	my $unzipped_body = $self->{body};
 
-	if((length($encoding) > 0) && defined($self->{body})) {
+	if(defined($self->{body})) {
 		my $range = $ENV{'Range'} ? $ENV{'Range'} : $ENV{'HTTP_RANGE'};
 
 		if($range && !$self->{cache}) {
@@ -335,7 +340,7 @@ sub DESTROY {
 				$unzipped_body = $self->{body};
 			}
 		}
-		if(length($self->{body}) >= MIN_GZIP_LEN) {
+		if((length($encoding) > 0) && (length($self->{body}) >= MIN_GZIP_LEN)) {
 			require Compress::Zlib;
 			Compress::Zlib->import;
 
@@ -366,6 +371,7 @@ sub DESTROY {
 				if(defined($self->{cobject})) {
 					$cache_hash = Storable::thaw($self->{cobject}->value());
 					$headers = $cache_hash->{'headers'};
+					$self->_set_content_type($headers);
 					@{$self->{o}} = ("X-FCGI-Buffer-$VERSION: Hit");
 					if($self->{info}) {
 						my $host_name = $self->{info}->host_name();
@@ -436,6 +442,9 @@ sub DESTROY {
 						}
 					}
 				}
+			}
+			if($self->{status} == 200) {
+				$encoding = $self->_should_gzip();
 				if($self->{send_body} && (length($encoding) > 0)) {
 					if(length($self->{body}) >= MIN_GZIP_LEN) {
 						require Compress::Zlib;
@@ -467,6 +476,9 @@ sub DESTROY {
 				}
 			}
 			if($ENV{'HTTP_IF_NONE_MATCH'} && $self->{send_body} && ($self->{status} != 304) && $self->{generate_304}) {
+				if($self->{logger}) {
+					$self->{logger}->debug("Compare $ENV{HTTP_IF_NONE_MATCH} with $self->{etag}");
+				}
 				if(defined($self->{etag}) && ($self->{etag} eq $ENV{'HTTP_IF_NONE_MATCH'}) && ($self->{status} == 200)) {
 					push @{$self->{o}}, "Status: 304 Not Modified";
 					$self->{send_body} = 0;
@@ -527,6 +539,9 @@ sub DESTROY {
 				if($self->{generate_etag} && defined($self->{etag})) {
 					$cache_hash->{'etag'} = $self->{etag};
 				}
+				# TODO: Support the Expires header
+				# if($headers !~ /^Expires: /m))) {
+				# }
 				$self->{cache}->set($key, Storable::freeze($cache_hash), $self->{cache_age});
 				if($self->{logger}) {
 					$self->{logger}->debug("store $key in the cache");
@@ -671,19 +686,31 @@ sub _check_modified_since {
 	}
 	my $s = HTTP::Date::str2time($$params{since});
 	if(!defined($s)) {
+		if($self->{logger}) {
+			$self->{logger}->info("$$params{since} is not a valid date");
+		}
 		# IF_MODIFIED_SINCE isn't a valid data
 		return;
 	}
 
 	my $age = $self->_my_age();
 	if(!defined($age)) {
+		if($self->{logger}) {
+			$self->{logger}->info("Can't determine my age");
+		}
 		return;
 	}
 	if($age > $s) {
+		if($self->{logger}) {
+			$self->{logger}->debug('_check_modified_since: script has been modified');
+		}
 		# Script has been updated so it may produce different output
 		return;
 	}
 
+	if($self->{logger}) {
+		$self->{logger}->debug("_check_modified_since: Compare $$params{modified} with $s");
+	}
 	if($$params{modified} <= $s) {
 		push @{$self->{o}}, "Status: 304 Not Modified";
 		$self->{status} = 304;
@@ -698,12 +725,12 @@ sub _optimise_content {
 	my $self = shift;
 
 	# Regexp::List - wow!
-	$self->{body} =~ s/(\s+|\r)\n|\n\+/\n/gs;
+	$self->{body} =~ s/(?^:(?:(?:\s+|\r)\n|\n(?:\s+|\+)))/\n/g;
 	# $self->{body} =~ s/\r\n/\n/gs;
 	# $self->{body} =~ s/\s+\n/\n/gs;
 	# $self->{body} =~ s/\n+/\n/gs;
-	$self->{body} =~ s/\<\/option\>\s\<option/\<\/option\>\<option/gis;
-	$self->{body} =~ s/\<\/div\>\s\<div/\<\/div\>\<div/gis;
+	# $self->{body} =~ s/\n\s+|\s+\n/\n/g;
+	$self->{body} =~ s/\<\/div\>\s+\<div/\<\/div\>\<div/gis;
 	# $self->{body} =~ s/\<\/p\>\s\<\/div/\<\/p\>\<\/div/gis;
 	# $self->{body} =~ s/\<div\>\s+/\<div\>/gis;	# Remove spaces after <div>
 	$self->{body} =~ s/(<div>\s+|\s+<div>)/<div>/gis;
@@ -716,8 +743,6 @@ sub _optimise_content {
 	$self->{body} =~ s/<body>\s+/<body>/is;
 	$self->{body} =~ s/\s+\<\/html/\<\/html/is;
 	$self->{body} =~ s/\s+\<\/body/\<\/body/is;
-	$self->{body} =~ s/\n\s+|\s+\n/\n/g;
-	$self->{body} =~ s/\t+/ /g;
 	$self->{body} =~ s/\s(\<.+?\>\s\<.+?\>)/$1/;
 	# $self->{body} =~ s/(\<.+?\>\s\<.+?\>)\s/$1/g;
 	$self->{body} =~ s/\<p\>\s/\<p\>/gi;
@@ -729,22 +754,22 @@ sub _optimise_content {
 	$self->{body} =~ s/\<br\s?\/?\>\s?\<p\>/\<p\>/gi;
 	$self->{body} =~ s/\<br\>\s/\<br\>/gi;
 	$self->{body} =~ s/\<br\s?\/\>\s/\<br \/\>/gi;
-	$self->{body} =~ s/ +/ /gs;	# Remove duplicate space, don't use \s+ it breaks JavaScript
+	$self->{body} =~ s/[ \t]+/ /gs;	# Remove duplicate space, don't use \s+ it breaks JavaScript
 	$self->{body} =~ s/\s\<p\>/\<p\>/gi;
 	$self->{body} =~ s/\s\<script/\<script/gi;
 	$self->{body} =~ s/(<script>\s|\s<script>)/<script>/gis;
 	$self->{body} =~ s/(<\/script>\s|\s<\/script>)/<\/script>/gis;
 	$self->{body} =~ s/\<td\>\s/\<td\>/gi;
-	$self->{body} =~ s/\s?\<a\shref="(.+?)"\>\s?/ <a href="$1">/gis;
+	$self->{body} =~ s/\s+\<a\shref="(.+?)"\>\s?/ <a href="$1">/gis;
 	$self->{body} =~ s/\s?<a\shref=\s"(.+?)"\>/ <a href="$1">/gis;
-	$self->{body} =~ s/(\s?<hr>\s|\s<hr>\s?)/<hr>/gis;
+	$self->{body} =~ s/\s+<\/a\>\s+/<\/a> /gis;
+	$self->{body} =~ s/(\s?<hr>\s+|\s+<hr>\s?)/<hr>/gis;
 	# $self->{body} =~ s/\s<hr>/<hr>/gis;
 	# $self->{body} =~ s/<hr>\s/<hr>/gis;
-	$self->{body} =~ s/<\/li>\s<li>/<\/li><li>/gis;
-	$self->{body} =~ s/<\/li>\s<\/ul>/<\/li><\/ul>/gis;
-	$self->{body} =~ s/<ul>\s<li>/<ul><li>/gis;
-
-	return $self->{body};
+	$self->{body} =~ s/<\/li>\+s<li>/<\/li><li>/gis;
+	$self->{body} =~ s/<\/li>\+s<\/ul>/<\/li><\/ul>/gis;
+	$self->{body} =~ s/<ul>\+s<li>/<ul><li>/gis;
+	$self->{body} =~ s/\<\/option\>\s+\<option/\<\/option\>\<option/gis;
 }
 
 # Create a key for the cache
@@ -762,9 +787,42 @@ sub _generate_key {
 	my $key = $self->{info}->browser_type() . '::' . $self->{info}->domain_name() . '::' . $self->{info}->script_name() . '::' . $self->{info}->as_string();
 	if($ENV{'HTTP_COOKIE'}) {
 		# Different states of the client are stored in different caches
-		$key .= '::' . $ENV{'HTTP_COOKIE'};
+		my $cookies = $ENV{'HTTP_COOKIE'};
+		# Don't put different Google Analytics in different caches, and anyway they
+		# would be wrong
+		foreach my $cookie(split(/;/, $ENV{'HTTP_COOKIE'})) {
+			unless($cookie =~ /^__utm[abcz]/) {
+				$key .= "::$cookie";
+			}
+		}
+	}
+
+	# Honour the Vary headers
+	my $headers = $self->{'headers'};
+	if($headers && ($headers =~ /^Vary: .*$/m)) {
+		if(defined($self->{logger})) {
+			$self->{logger}->debug('Found Vary header');
+		}
+		foreach my $h1(split(/\r?\n/, $headers)) {
+			my ($h1_name, $h1_value) = split /\:\s*/, $h1, 2;
+			if(lc($h1_name) eq 'vary') {
+				foreach my $h2(split(/\r?\n/, $headers)) {
+					my ($h2_name, $h2_value) = split /\:\s*/, $h2, 2;
+					if($h2_name eq $h1_value) {
+						$key .= '::' . $h2_value;
+						last;
+					}
+				}
+			}
+		}
 	}
 	$key =~ s/\//::/g;
+	$key =~ s/::::/::/g;
+	$key =~ s/::$//;
+	if(defined($self->{logger})) {
+		$self->{logger}->trace("Returning $key");
+	}
+	$self->{cache_key} = $key;
 	return $key;
 }
 
@@ -926,6 +984,7 @@ sub can_cache {
 		if(defined($self->{logger})) {
 			$self->{logger}->debug("cache_control = $control");
 		}
+		# TODO: check Authorization header not present
 		if(($control eq 'no-store') ||
 		   ($control eq 'no-cache') ||
 		   ($control eq 'max-age=0') ||
@@ -1090,6 +1149,21 @@ sub _should_gzip {
 	return '';
 }
 
+sub _set_content_type {
+	my $self = shift;
+	my $headers = shift;
+
+	foreach my $header (split(/\r?\n/, $headers)) {
+		my ($header_name, $header_value) = split /\:\s*/, $header, 2;
+		if (lc($header_name) eq 'content-type') {
+			my @content_type;
+			@content_type = split /\//, $header_value, 2;
+			$self->{content_type} = \@content_type;
+			return;
+		}
+	}
+}
+
 =head1 AUTHOR
 
 Nigel Horne, C<< <njh at bandsman.co.uk> >>
